<?php

namespace Drupal\tome\Commands;

use Drupal\Core\Entity\ContentEntityInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Session\AccountSwitcherInterface;
use Drupal\tome\ConcurrentProcessTrait;
use Drupal\tome\AccountSwitcherTrait;
use Drupal\tome\Event\TomeEvents;
use Drupal\tome\ExporterInterface;
use Drush\Commands\DrushCommands;
use Symfony\Component\EventDispatcher\Event;
use Symfony\Component\EventDispatcher\EventDispatcherInterface;

/**
 * Contains export commands for tome.
 */
class ExportCommands extends DrushCommands {

  use ConcurrentProcessTrait;
  use AccountSwitcherTrait;

  /**
   * The default number of processes to invoke.
   *
   * @todo Increase this once deadlocks do not occur in SQLite. :-(
   *
   * @var int
   */
  const PROCESS_COUNT = 1;

  /**
   * The default number of entities to import in each process.
   *
   * @var int
   */
  const ENTITY_COUNT = 20;

  /**
   * The exporter.
   *
   * @var \Drupal\tome\ExporterInterface
   */
  protected $exporter;

  /**
   * The entity type manager.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected $entityTypeManager;

  /**
   * The event dispatcher.
   *
   * @var \Symfony\Component\EventDispatcher\EventDispatcherInterface
   */
  protected $eventDispatcher;

  /**
   * Constructs an ExportCommands instance.
   *
   * @param \Drupal\tome\ExporterInterface $exporter
   *   The exporter.
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entity_type_manager
   *   The entity type manager.
   * @param \Drupal\Core\Session\AccountSwitcherInterface $account_switcher
   *   The account switcher.
   * @param \Symfony\Component\EventDispatcher\EventDispatcherInterface $event_dispatcher
   *   The event dispatcher.
   */
  public function __construct(ExporterInterface $exporter, EntityTypeManagerInterface $entity_type_manager, AccountSwitcherInterface $account_switcher, EventDispatcherInterface $event_dispatcher) {
    $this->exporter = $exporter;
    $this->entityTypeManager = $entity_type_manager;
    $this->accountSwitcher = $account_switcher;
    $this->eventDispatcher = $event_dispatcher;
  }

  /**
   * Exports given content.
   *
   * @param string $chunk
   *   A comma separated list of ID pairs in the format entity_type_id:id.
   *
   * @command tome:export-content
   *
   * @return int|null
   *   The status code, if the command did not complete successfully.
   */
  public function exportContent($chunk) {
    $this->switchToAdmin();
    $id_pairs = explode(',', $chunk);
    $storages = [];
    foreach ($id_pairs as $id_pair) {
      list($entity_type_id, $id) = explode(':', $id_pair);
      $entity_type = $this->entityTypeManager->getDefinition($entity_type_id);
      if (!$entity_type) {
        $this->io()->error("The entity type $entity_type_id does not exist.");
        return 1;
      }
      if (!isset($storages[$entity_type_id])) {
        $storages[$entity_type_id] = $this->entityTypeManager->getStorage($entity_type_id);
      }
      $entity = $storages[$entity_type_id]->load($id);
      if (!$entity) {
        $this->io()->error("No entity found for $id_pair.");
        return 1;
      }
      if (!($entity instanceof ContentEntityInterface)) {
        $this->io()->error("$id_pair is not a content entity.");
        return 1;
      }
      foreach ($entity->getTranslationLanguages() as $language) {
        $this->exporter->exportContent($entity->getTranslation($language->getId()));
      }
    }
    $this->switchBack();
  }

  /**
   * Exports all config, content, and files.
   *
   * @param array $options
   *   An array of Drush options.
   *
   * @command tome:export
   *
   * @return int|null
   *   The status code, if the command did not complete successfully.
   */
  public function exportAll(array $options = ['process-count' => self::PROCESS_COUNT, 'entity-count' => self::ENTITY_COUNT]) {
    if (!$this->io()->confirm('The files in your export directory will be deleted and replaced.', FALSE)) {
      return 0;
    }
    drush_invoke_process('@self', 'config-export', [], ['yes' => TRUE]);

    $this->exporter->deleteExportDirectories();
    $entities = $this->exporter->getContentToExport();
    $id_pairs = [];
    $commands = [];
    foreach ($entities as $entity_type_id => $ids) {
      foreach ($ids as $id) {
        $id_pairs[] = "$entity_type_id:$id";
      }
    }
    foreach (array_chunk($id_pairs, $options['entity-count']) as $chunk) {
      $commands[] = drush_build_drush_command() . ' tome:export-content ' . implode(',', $chunk);
    }
    $collected_errors = $this->runCommands($commands, $options['process-count']);
    if (!empty($collected_errors)) {
      $this->io()->error('Errors encountered when exporting content:');
      $this->displayErrors($collected_errors);
      return 1;
    }

    $this->eventDispatcher->dispatch(TomeEvents::EXPORT_ALL, new Event());

    $this->io()->success('Exported config, content, and files.');
  }

}
