<?php

namespace Drupal\tome\Commands;

use Drupal\Core\Entity\ContentEntityInterface;
use Drupal\Core\Entity\ContentEntityTypeInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Session\AccountSwitcherInterface;
use Drupal\tome\ConcurrentProcessTrait;
use Drupal\tome\AccountSwitcherTrait;
use Drupal\tome\ImporterInterface;
use Drupal\tome\TomeHelper;
use Drush\Commands\DrushCommands;

/**
 * Contains export commands for tome.
 */
class ImportCommands extends DrushCommands {

  use ConcurrentProcessTrait;
  use AccountSwitcherTrait;

  /**
   * The default number of processes to invoke.
   *
   * @todo Increase this once deadlocks do not occur in SQLite. :-(
   *
   * @var int
   */
  const PROCESS_COUNT = 1;

  /**
   * The default number of entities to import in each process.
   *
   * @var int
   */
  const ENTITY_COUNT = 20;

  /**
   * The importer.
   *
   * @var \Drupal\tome\ImporterInterface
   */
  protected $importer;

  /**
   * The entity type manager.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected $entityTypeManager;

  /**
   * Constructs an ExportCommands instance.
   *
   * @param \Drupal\tome\ImporterInterface $importer
   *   The importer.
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entity_type_manager
   *   The entity type manager.
   * @param \Drupal\Core\Session\AccountSwitcherInterface $account_switcher
   *   The account switcher.
   */
  public function __construct(ImporterInterface $importer, EntityTypeManagerInterface $entity_type_manager, AccountSwitcherInterface $account_switcher) {
    $this->importer = $importer;
    $this->entityTypeManager = $entity_type_manager;
    $this->accountSwitcher = $account_switcher;
  }

  /**
   * Imports given content.
   *
   * @param string $names
   *   A comma separated list of content names.
   *
   * @command tome:import-content
   */
  public function importContent($names) {
    $this->switchToAdmin();
    $names = explode(',', $names);
    foreach ($names as $name) {
      list($entity_type_id, $uuid, $langcode) = TomeHelper::getPartsFromContentName($name);
      $this->importer->importContent($entity_type_id, $uuid, $langcode);
    }
    $this->switchBack();
  }

  /**
   * Imports all config, content, and files.
   *
   * @param array $options
   *   An array of Drush options.
   *
   * @command tome:import
   * @option process-count Limits the number of pages that will process at the same time.
   * @option entity-count Limits the number of number of entities to import in each process.
   *
   * @return int|null
   *   The status code, if the command did not complete successfully.
   */
  public function importAll(array $options = ['process-count' => self::PROCESS_COUNT, 'entity-count' => self::ENTITY_COUNT]) {
    if (!$this->io()->confirm('Your local site\'s config, content, and files will be deleted and replaced.', FALSE)) {
      return;
    }

    \Drupal::service('config.installer')->setSyncing(TRUE);
    $this->importer->isImporting(TRUE);
    foreach ($this->entityTypeManager->getDefinitions() as $entity_type) {
      if ($entity_type instanceof ContentEntityTypeInterface) {
        foreach ($this->entityTypeManager->getStorage($entity_type->id())->loadMultiple() as $entity) {
          if ($entity instanceof ContentEntityInterface) {
            $entity->delete();
          }
        }
      }
    }
    $this->importer->isImporting(FALSE);
    \Drupal::service('config.installer')->setSyncing(FALSE);

    /** @var \Drupal\Core\Config\StorageInterface $source_storage */
    $source_storage = \Drupal::service('config.storage.sync');
    \Drupal::configFactory()->getEditable('system.site')->setData($source_storage->read('system.site'))->save(TRUE);
    drush_invoke_process('@self', 'config-import', [], ['yes' => TRUE]);

    $chunked_names = $this->importer->getChunkedNames();
    foreach ($chunked_names as $chunk) {
      $commands = [];
      foreach (array_chunk($chunk, $options['entity-count']) as $names) {
        $commands[] = drush_build_drush_command() . ' tome:import-content ' . implode(',', $names);
      }
      $collected_errors = $this->runCommands($commands, $options['process-count']);
      if (!empty($collected_errors)) {
        $this->io()->error('Errors encountered when importing content:');
        $this->displayErrors($collected_errors);
        return 1;
      }
    }

    $this->importer->importFiles();

    drush_invoke_process('@self', 'cache-rebuild');

    // We need a fresh container to properly invoke events.
    drush_invoke_process('@self', 'php:eval', ['\Drupal::service("event_dispatcher")->dispatch(\Drupal\tome\Event\TomeEvents::IMPORT_ALL, new \Symfony\Component\EventDispatcher\Event())']);

    $this->io()->success('Imported config, content, and files.');
  }

}
