<?php

namespace Drupal\tome\Commands;

use Drupal\tome\ConcurrentProcessTrait;
use Drupal\tome\StaticCache;
use Drupal\tome\StaticGeneratorInterface;
use Drush\Commands\DrushCommands;
use Drush\Drush;
use Drush\Exec\ExecTrait;
use Symfony\Component\Process\Process;

/**
 * Contains static site commands for tome.
 */
class StaticCommands extends DrushCommands {

  use ConcurrentProcessTrait;
  use ExecTrait;

  /**
   * The default number of processes to invoke.
   *
   * @var int
   */
  const PROCESS_COUNT = 5;

  /**
   * The static service.
   *
   * @var \Drupal\tome\StaticGeneratorInterface
   */
  protected $static;

  /**
   * The static cache.
   *
   * @var \Drupal\tome\StaticCache
   */
  protected $cache;

  /**
   * Constructs a StaticCommands instance.
   *
   * @param \Drupal\tome\StaticGeneratorInterface $static
   *   The static service.
   * @param \Drupal\tome\StaticCache $cache
   *   The static cache.
   */
  public function __construct(StaticGeneratorInterface $static, StaticCache $cache) {
    $this->static = $static;
    $this->cache = $cache;
  }

  /**
   * Exports all static HTML for the site.
   *
   * @param array $options
   *   An array of Drush options.
   *
   * @command tome:static
   * @option process-count Limits the number of pages that will process at the same time.
   * @option run-server If a local HTTP server should be started after the export. Defaults to FALSE.
   * @option port The port to run the server on. Defaults to 8889.
   * @option ignore-warnings If configuration warnings should be shown. Defaults to FALSE.
   * @option partial If only a partial export is desired. Defaults to FALSE.
   * @option path-pattern If you only want to export a specific paths based on pattern.
   */
  public function export(array $options = [
    'process-count' => self::PROCESS_COUNT,
    'run-server' => FALSE,
    'port' => 8889,
    'ignore-warnings' => FALSE,
    'partial' => FALSE,
    'path-pattern' => '',
  ]) {
    $this->cache->deleteAll($options['partial']);

    $warnings = $this->getWarnings($options);
    if (!$options['ignore-warnings'] && $warnings) {
      $warnings[] = 'To suppress these messages, pass the --ignore-warnings option.';
      $this->io()->comment($warnings);
    }

    // @todo Is this a Drush bug?
    if (base_path() !== '/') {
      \Drupal::urlGenerator()->getContext()->setBaseUrl(rtrim(base_path(), '/'));
    }

    $target_dir = $this->static->getStaticDirectory();
    if (!$options['partial'] && file_exists($target_dir)) {
      file_unmanaged_delete_recursive($target_dir);
    }
    file_prepare_directory($target_dir, FILE_CREATE_DIRECTORY);

    $paths = $this->static->getPaths($options['partial']);

    // Add a filter to export only pattern based paths.
    if ($options['path-pattern'] !== '') {
      $paths = preg_filter($options['path-pattern'], '$0', $paths);
    }

    $this->io()->writeln('Generating static HTML...');
    $this->exportPaths($paths, [], $options['process-count'], TRUE, $options['uri'], $options['partial']);
    $this->io()->success('Exported static HTML and related assets.');

    $this->cache->commit();

    if ($options['run-server']) {
      $url = '127.0.0.1:' . $options['port'];
      $this->startBrowser('http://' . $url . base_path(), 2);
      passthru('cd ' . escapeshellarg($target_dir) . ' && ' . Drush::config()->get('php', 'php') . ' -S ' . escapeshellarg($url));
    }
  }

  /**
   * Exports a given path.
   *
   * @param string $path
   *   A path to export, or an entity to load and export in the format
   *   "_entity:entity_type_id:langcode:entity_id".
   * @param array $options
   *   An array of Drush options.
   *
   * @command tome:static-export-path
   * @option process-count Limits the number of pages that will process at the same time.
   * @option show-progress Whether or not a progress bar should be shown. Defaults to FALSE.
   * @option return-json Whether or not paths that need invoking should be returned as JSON. Defaults to FALSE.
   * @option partial If only a partial export is desired. Defaults to FALSE.
   */
  public function exportPath($path, array $options = [
    'process-count' => self::PROCESS_COUNT,
    'show-progress' => FALSE,
    'return-json' => FALSE,
    'partial' => FALSE,
  ]) {
    $invoke_paths = $this->static->requestPath($path, $options['partial']);
    if ($options['return-json']) {
      $this->io()->write(json_encode($invoke_paths, JSON_PRETTY_PRINT));
    }
    else {
      $this->exportPaths($invoke_paths, [$path], $options['process-count'], $options['show-progress'], $options['uri'], $options['partial']);
    }
  }

  /**
   * Exports the given paths to the static directory.
   *
   * @param string[] $paths
   *   An array of paths.
   * @param array $old_paths
   *   An array of paths that have already been processed.
   * @param int $process_count
   *   The number of processes to invoke.
   * @param bool $show_progress
   *   Whether or not a progress bar should be shown.
   * @param string $uri
   *   The URI of the site, probably passed by -l or --uri.
   * @param bool $partial
   *   Whether or not this is a partial build.
   */
  protected function exportPaths(array $paths, array $old_paths, $process_count, $show_progress, $uri, $partial) {
    $paths = $this->static->exportPaths($paths);

    if (empty($paths)) {
      return;
    }

    if ($this->io()->isVerbose()) {
      $this->io()->writeln('Exporting paths:');
      $this->io()->listing($paths);
    }

    $commands = [];
    foreach ($paths as $path) {
      $command = drush_build_drush_command() . ' tome:static-export-path --return-json ' . escapeshellarg($path) . ' --process-count=' . escapeshellarg($process_count) . ' --uri=' . escapeshellarg($uri);
      if ($partial) {
        $command .= ' --partial';
      }
      $commands[] = $command;
    }

    $show_progress && $this->io()->progressStart(count($paths));

    $invoke_paths = [];
    $collected_errors = $this->runCommands($commands, $process_count, function (Process $process) use ($show_progress, &$invoke_paths) {
      $show_progress && $this->io()->progressAdvance();
      $output = $process->getOutput();
      if (!empty($output) && $json = json_decode($output, TRUE)) {
        $invoke_paths = array_merge($invoke_paths, $json);
      }
    });

    $invoke_paths = array_diff($invoke_paths, $old_paths);
    $old_paths = array_merge($old_paths, $invoke_paths);

    $show_progress && $this->io()->progressFinish();
    if (!empty($collected_errors)) {
      $this->displayErrors($collected_errors);
    }
    if (count($invoke_paths)) {
      $this->io()->writeln('Processing related assets and paths...');
      $this->exportPaths($invoke_paths, $old_paths, $process_count, $show_progress, $uri, $partial);
    }
  }

  /**
   * Collects warnings to help users correct issues in rendered HTML.
   *
   * @param array $options
   *   An array of options passed to the tome:static command.
   *
   * @return array
   *   An array of warning messages to display to the user.
   */
  protected function getWarnings(array $options) {
    $warnings = [];
    if (empty($options['uri'])) {
      $warnings[] = 'No "--uri" option provided. This could lead to invalid absolute URLs. To resolve, pass the "--uri" option.';
    }
    $performance_config = \Drupal::config('system.performance');
    if (!$performance_config->get('css.preprocess') || !$performance_config->get('js.preprocess')) {
      if (!$performance_config->get('css.preprocess') && !$performance_config->get('js.preprocess')) {
        $message = 'CSS and JS preprocessing is disabled.';
      }
      elseif (!$performance_config->get('css.preprocess')) {
        $message = 'CSS preprocessing is disabled.';
      }
      else {
        $message = 'JS preprocessing is disabled.';
      }
      $warnings[] = $message . ' This could lead to performance issues. To resolve, visit /admin/config/development/performance.';
    }
    $twig_config = \Drupal::getContainer()->getParameter('twig.config');
    if ($twig_config['debug'] || !$twig_config['cache']) {
      if ($twig_config['debug'] && !$twig_config['cache']) {
        $message = 'Twig debugging is enabled and caching is disabled.';
      }
      elseif ($twig_config['debug']) {
        $message = 'Twig debugging is enabled.';
      }
      else {
        $message = 'Twig caching is disabled.';
      }
      $warnings[] = $message . ' This could lead to performance issues. To resolve, edit the "twig.config" parameter in the "sites/*/services.yml" file.';
    }
    return $warnings;
  }

}
