<?php

namespace Drupal\tome;

use Symfony\Component\Process\Process;

/**
 * Shared methods for handling concurrency.
 */
trait ConcurrentProcessTrait {

  /**
   * Runs commands with concurrency.
   *
   * @param string[] $commands
   *   An array of commands to execute.
   * @param int $concurrency
   *   The number of concurrent processes to execute.
   * @param callback|\Closure $callback
   *   (optional) A callback to invoke for each completed callback.
   *
   * @return array
   *   An array of errors encountered when running commands.
   */
  protected function runCommands(array $commands, $concurrency, $callback = NULL) {
    $current_processes = [];
    $collected_errors = [];

    $filter_callback = function (Process $current_process) use (&$collected_errors, $callback) {
      $is_running = $current_process->isRunning();
      $command = $current_process->getCommandLine();
      if (!$is_running) {
        $error_output = $current_process->getErrorOutput();
        if ($error_output) {
          $collected_errors[] = "Error when running \"{$command}\":\n  $error_output";
        }
        if ($callback) {
          call_user_func($callback, $current_process);
        }
      }
      return $is_running;
    };

    while ($commands || $current_processes) {
      $current_processes = array_filter($current_processes, $filter_callback);
      if ($commands && count($current_processes) < $concurrency) {
        $command = array_shift($commands);
        $process = new Process($command);
        $process->start();
        $current_processes[] = $process;
      }
    }

    return $collected_errors;
  }

  /**
   * Displays errors using the IO component.
   *
   * @param string[] $collected_errors
   *   An array of error messages to display.
   */
  protected function displayErrors(array $collected_errors) {
    foreach ($collected_errors as $error) {
      $this->io()->error($error);
    }
  }

}
