<?php

namespace Drupal\tome;

use Drupal\Core\Cache\CacheableResponseInterface;
use Drupal\Core\Config\FileStorage;
use Drupal\Core\Config\StorageInterface;
use Drupal\Core\KeyValueStore\KeyValueFactoryInterface;
use Drupal\Core\Site\Settings;
use Symfony\Component\HttpFoundation\Response;

/**
 * A class that checks and maintains metadata about static cache.
 *
 * @internal
 */
class StaticCache {

  /**
   * The config storage.
   *
   * @var \Drupal\Core\Config\StorageInterface
   */
  protected $configStorage;

  /**
   * The content storage.
   *
   * @var \Drupal\Core\Config\StorageInterface
   */
  protected $contentStorage;

  /**
   * The key value store.
   *
   * @var \Drupal\Core\KeyValueStore\KeyValueStoreInterface
   */
  protected $keyValue;

  /**
   * The cache metadata.
   *
   * @var array
   */
  protected $cache;

  /**
   * A map of cache tag names to file paths.
   *
   * @var array
   */
  protected $tagFileMap;

  /**
   * Constructs a StaticCacheChecker object.
   *
   * @param \Drupal\Core\KeyValueStore\KeyValueFactoryInterface $key_value_factory
   *   The key value store.
   * @param \Drupal\Core\Config\StorageInterface $config_storage
   *   The config storage.
   * @param \Drupal\Core\Config\StorageInterface $content_storage
   *   The content storage.
   */
  public function __construct(KeyValueFactoryInterface $key_value_factory, StorageInterface $config_storage, StorageInterface $content_storage) {
    $this->keyValue = $key_value_factory->get('tome_static_cache');
    $this->configStorage = $config_storage;
    $this->contentStorage = $content_storage;
  }

  /**
   * Determines if a request is cached.
   *
   * @param string $path
   *   The request path.
   *
   * @return bool
   *   TRUE if the request is cached, FALSE otherwise.
   */
  public function isCached($path) {
    $request_cache = $this->getCacheMetadata($path);
    if (!$request_cache) {
      return FALSE;
    }
    $never_cache = Settings::get('tome_static_cache_exclude', []);
    if (is_array($never_cache)) {
      foreach ($never_cache as $pattern) {
        if ($path === $pattern || @preg_match($pattern, $path)) {
          return FALSE;
        }
      }
    }
    foreach ($request_cache['files'] as $file => $hash) {
      if (!file_exists($file) || sha1_file($file) !== $hash) {
        return FALSE;
      }
    }
    return TRUE;
  }

  /**
   * Adds cache metadata for a request/response pair.
   *
   * Once all static pages are generated, the commit method should be run to
   * ensure the cache metadata is included in the static build.
   *
   * @param string $path
   *   The request path.
   * @param \Symfony\Component\HttpFoundation\Response $response
   *   The response.
   */
  public function add($path, Response $response) {
    if ($response instanceof CacheableResponseInterface) {
      $map = $this->getTagFileMap();
      $files = [];
      foreach ($response->getCacheableMetadata()->getCacheTags() as $cache_tag) {
        if (isset($map[$cache_tag]) && file_exists($map[$cache_tag])) {
          $files[$map[$cache_tag]] = sha1_file($map[$cache_tag]);
        }
      }
      $this->keyValue->set($path, [
        'files' => $files,
      ]);
    }
  }

  /**
   * Removes all cache metadata from the database.
   *
   * If you're starting a new static build, you should call this to ensure that
   * the database does not include any old cache metadata.
   *
   * @param bool $reload
   *   If the cache metadata should be reloaded from the filesystem.
   */
  public function deleteAll($reload = FALSE) {
    $this->keyValue->deleteAll();
    $cache_filename = $this->getCacheFilename();
    if ($reload && file_exists($cache_filename)) {
      $this->cache = json_decode(file_get_contents($cache_filename), TRUE);
      $this->keyValue->setMultiple($this->cache);
    }
  }

  /**
   * Writes the static metadata to the static build directory.
   */
  public function commit() {
    $cache_filename = $this->getCacheFilename();
    $directory = dirname($cache_filename);
    file_prepare_directory($directory, FILE_CREATE_DIRECTORY);
    $cache = $this->keyValue->getAll();
    file_put_contents($cache_filename, json_encode($cache, JSON_PRETTY_PRINT), FILE_EXISTS_REPLACE);
    $this->cache = $cache;
  }

  /**
   * Gets the cache metadata from the filesystem.
   *
   * @param string $path
   *   The request path.
   *
   * @return array|bool
   *   The cache metadata, or FALSE if the cache does not exist.
   */
  protected function getCacheMetadata($path) {
    if (!isset($this->cache)) {
      $cache_filename = $this->getCacheFilename();
      if (!file_exists($cache_filename)) {
        return [];
      }
      $this->cache = json_decode(file_get_contents($cache_filename), TRUE);
    }
    return isset($this->cache[$path]) ? $this->cache[$path] : FALSE;
  }

  /**
   * Gets the cache filename.
   *
   * @return string
   *   The filename.
   */
  protected function getCacheFilename() {
    return Settings::get('tome_static_directory', '../html') . '/.cache.json';
  }

  /**
   * Returns a map of cache tags to file paths.
   *
   * @return string[]
   *   An array mapping cache tag names to file paths.
   */
  protected function getTagFileMap() {
    if (!isset($this->tagFileMap)) {
      $this->tagFileMap = [];
      if ($this->configStorage instanceof FileStorage) {
        foreach ($this->configStorage->listAll() as $name) {
          $this->tagFileMap["config:$name"] = $this->configStorage->getFilePath($name);
        }
      }
      if ($this->contentStorage instanceof FileStorage) {
        foreach ($this->contentStorage->listAll() as $name) {
          list($entity_type_id, $uuid) = TomeHelper::getPartsFromContentName($name);
          $this->tagFileMap["$entity_type_id:$uuid"] = $this->contentStorage->getFilePath($name);
        }
      }
    }
    return $this->tagFileMap;
  }

}
