<?php

namespace Drupal\tome;

use Drupal\Component\Utility\UrlHelper;
use Drupal\Core\Site\Settings;
use Drupal\tome\Event\CollectPathsEvent;
use Drupal\tome\Event\PathPlaceholderEvent;
use Drupal\tome\Event\TomeEvents;
use Symfony\Component\EventDispatcher\EventDispatcherInterface;
use Symfony\Component\HttpFoundation\BinaryFileResponse;
use Symfony\Component\HttpFoundation\Request;
use Symfony\Component\HttpFoundation\RequestStack;
use Symfony\Component\HttpKernel\HttpKernelInterface;

/**
 * Handles static site generation.
 */
class StaticGenerator implements StaticGeneratorInterface {

  /**
   * The HTTP kernel.
   *
   * @var \Symfony\Component\HttpKernel\HttpKernelInterface
   */
  protected $httpKernel;

  /**
   * The request stack.
   *
   * @var \Symfony\Component\HttpFoundation\Request
   */
  protected $currentRequest;

  /**
   * The event dispatcher.
   *
   * @var \Symfony\Component\EventDispatcher\EventDispatcherInterface
   */
  protected $eventDispatcher;

  /**
   * The static cache.
   *
   * @var \Drupal\tome\StaticCache
   */
  protected $cache;

  /**
   * Creates a StaticGenerator object.
   *
   * @param \Symfony\Component\HttpKernel\HttpKernelInterface $http_kernel
   *   The HTTP kernel.
   * @param \Symfony\Component\HttpFoundation\RequestStack $request_stack
   *   The request stack.
   * @param \Symfony\Component\EventDispatcher\EventDispatcherInterface $event_dispatcher
   *   The event dispatcher.
   * @param \Drupal\tome\StaticCache $cache
   *   The static cache.
   */
  public function __construct(HttpKernelInterface $http_kernel, RequestStack $request_stack, EventDispatcherInterface $event_dispatcher, StaticCache $cache) {
    $this->httpKernel = $http_kernel;
    $this->currentRequest = $request_stack->getCurrentRequest();
    $this->eventDispatcher = $event_dispatcher;
    $this->cache = $cache;
  }

  /**
   * {@inheritdoc}
   */
  public function getPaths($partial = FALSE) {
    $event = new CollectPathsEvent([]);
    $this->eventDispatcher->dispatch(TomeEvents::COLLECT_PATHS, $event);
    $paths = array_unique($event->getPaths());
    $paths = array_values(array_diff($paths, $this->getExcludedPaths()));

    if ($partial) {
      $paths = array_filter($paths, function ($path) {
        return !$this->cache->isCached($path);
      });
    }

    return $paths;
  }

  /**
   * {@inheritdoc}
   */
  public function requestPath($path, $partial = FALSE) {
    $invoke_paths = [];

    $original_path = $path;

    $event = new PathPlaceholderEvent($path);
    $this->eventDispatcher->dispatch(TomeEvents::PATH_PLACEHOLDER, $event);

    if ($event->isInvalid()) {
      return [];
    }

    $path = $event->getPath();

    $request = Request::create($path, 'GET', [], [], [], $this->currentRequest->server->all());
    $response = $this->httpKernel->handle($request, HttpKernelInterface::MASTER_REQUEST);
    if ($response->getStatusCode() === 200) {
      $destination = $this->getDestination($path);
      $directory = dirname($destination);
      file_prepare_directory($directory, FILE_CREATE_DIRECTORY);
      // This is probably an image style derivative.
      if ($response instanceof BinaryFileResponse) {
        $file_path = $response->getFile()->getPathname();
        $this->copyPath($file_path, $destination);
      }
      else {
        $content = $response->getContent();
        if (strpos($response->headers->get('Content-Type'), 'text/html') === 0) {
          $invoke_paths = array_merge($invoke_paths, $this->handlePagers($content, $path));
          $invoke_paths = array_merge($invoke_paths, $this->getHtmlAssets($content, $path));
        }
        file_put_contents($destination, $content);
      }

      $this->cache->add($original_path, $response);
    }

    if ($partial) {
      $invoke_paths = array_filter($invoke_paths, function ($invoke_path) {
        return !$this->cache->isCached($invoke_path);
      });
    }

    return $invoke_paths;
  }

  /**
   * {@inheritdoc}
   */
  public function exportPaths(array $paths) {
    $paths = array_diff($paths, $this->getExcludedPaths());
    $paths = array_values(array_unique($paths));

    $invoke_paths = [];

    foreach ($paths as $path) {
      $components = parse_url($path);
      if (UrlHelper::isExternal($path) && isset($components['host']) && UrlHelper::externalIsLocal($path, $this->currentRequest->getSchemeAndHttpHost())) {
        $path = $components['path'];
        if (!empty($components['query'])) {
          $path .= '?' . $components['query'];
        }
        if (!empty($components['fragment'])) {
          $path .= '#' . $components['fragment'];
        }
      }
      $destination = $this->getDestination($path);
      if (UrlHelper::isExternal($path)) {
        continue;
      }

      $sanitized_path = $this->sanitizePath($path);
      if ($this->copyPath($sanitized_path, $destination)) {
        if (pathinfo($destination, PATHINFO_EXTENSION) === 'css') {
          $css_assets = $this->getCssAssets(file_get_contents($destination), $sanitized_path);
          $invoke_paths = array_merge($invoke_paths, $this->exportPaths($css_assets));
        }
      }
      else {
        $invoke_paths[] = $path;
      }
    }

    return $invoke_paths;
  }

  /**
   * {@inheritdoc}
   */
  public function getStaticDirectory() {
    return Settings::get('tome_static_directory', '../html');
  }

  /**
   * Finds assets for the given CSS content.
   *
   * @param string $content
   *   A CSS string.
   * @param string $root
   *   A root path to resolve relative paths.
   *
   * @return array
   *   An array of paths found in the given CSS string.
   */
  protected function getCssAssets($content, $root) {
    $paths = [];
    // Regex copied from the Static module from Drupal 7.
    // Credit to Randall Knutson and Michael Vanetta.
    $matches = [];
    preg_match_all('/url\(\s*[\'"]?(?!(?:data)+:)([^\'")]+)[\'"]?\s*\)/i', $content, $matches);
    if (isset($matches[1])) {
      $paths = $matches[1];
    }
    $paths = $this->getRealPaths($paths, $root);
    return $paths;
  }

  /**
   * Turns relative paths into absolute paths.
   *
   * Useful specifically for CSS's url().
   *
   * @param array $paths
   *   An array of paths to convert.
   * @param string $root
   *   A root path to resolve relative paths.
   *
   * @return array
   *   An array of converted paths.
   */
  protected function getRealPaths(array $paths, $root) {
    $root_dir = dirname($this->sanitizePath($root));
    foreach ($paths as &$path) {
      if (strpos($path, '../') !== FALSE) {
        $path = $root_dir . '/' . $path;
      }
    }
    return $paths;
  }

  /**
   * Converts pager links to distinct URLs and returns paths for processing.
   *
   * @param string &$content
   *   An HTML string.
   * @param string $path
   *   The current path.
   *
   * @return array
   *   An array of pager paths found in the given HTML string..
   */
  protected function handlePagers(&$content, $path) {
    $paths = [];
    $document = new \DOMDocument();
    @$document->loadHTML($content);
    $xpath = new \DOMXPath($document);
    /** @var \DOMElement $node */
    foreach ($xpath->query('//a[contains(@href,"?page=")]') as $node) {
      $original_href = $node->getAttribute('href');
      if ($original_href[0] === '?') {
        $new_href = strtok($path, '?') . $original_href;
      }
      else {
        $new_href = $original_href;
      }
      $paths[] = $new_href;
      $new_href = $this->replacePagerUrl($new_href);
      $content = str_replace($original_href, $new_href, $content);
    }
    return $paths;
  }

  /**
   * Replaces pager params in a URL with an appropriate path.
   *
   * @param string $url
   *   A URL that may contain a pager (?page) query param.
   *
   * @return string
   *   A modified URL.
   */
  protected function replacePagerUrl($url) {
    parse_str(parse_url($url, PHP_URL_QUERY), $query);
    if ($query && isset($query['page']) && is_numeric($query['page'])) {
      $base_path = preg_replace('/\?.*/', '', $url);
      if ($query['page'] != 0) {
        if ($base_path === '/') {
          $base_path = '';
        }
        $url = $base_path . '/page/' . ($query['page'] + 1);
      }
      else {
        $url = !empty($base_path) ? $base_path : '/';
      }
    }
    return $url;
  }

  /**
   * Finds and exports assets for the given HTML content.
   *
   * @param string $content
   *   An HTML string.
   * @param string $root
   *   A root path to resolve relative paths.
   *
   * @return array
   *   An array of paths found in the given HTML string.
   */
  protected function getHtmlAssets($content, $root) {
    $paths = [];
    $document = new \DOMDocument();
    @$document->loadHTML($content);
    $xpath = new \DOMXPath($document);
    /** @var \DOMElement $image */
    foreach ($document->getElementsByTagName('img') as $image) {
      if ($image->hasAttribute('src')) {
        $paths[] = $image->getAttribute('src');
      }
      if ($image->hasAttribute('srcset')) {
        $srcset = $image->getAttribute('srcset');
        $sources = explode(' ', preg_replace('/ [^ ]+(,|$)/', '', $srcset));
        foreach ($sources as $src) {
          $paths[] = $src;
        }
      }
    }
    $rels = [
      'stylesheet',
      'shortcut icon',
      'image_src',
    ];
    /** @var \DOMElement $node */
    foreach ($document->getElementsByTagName('link') as $node) {
      if (in_array($node->getAttribute('rel'), $rels, TRUE) && $node->hasAttribute('href')) {
        $paths[] = $node->getAttribute('href');
      }
    }
    /** @var \DOMElement $node */
    foreach ($document->getElementsByTagName('script') as $node) {
      if ($node->hasAttribute('src')) {
        $paths[] = $node->getAttribute('src');
      }
    }
    /** @var \DOMElement $node */
    foreach ($document->getElementsByTagName('style') as $node) {
      $paths = array_merge($paths, $this->getCssAssets($node->textContent, $root));
    }
    foreach ($xpath->query('//*[@style]') as $node) {
      $paths = array_merge($paths, $this->getCssAssets($node->getAttribute('style'), $root));
    }
    return $paths;
  }

  /**
   * Returns a destination for saving a given path.
   *
   * @param string $path
   *   The path.
   *
   * @return string
   *   The destination.
   */
  protected function getDestination($path) {
    $path = $this->replacePagerUrl($path);
    $path = urldecode($path);
    if (empty(pathinfo($path, PATHINFO_EXTENSION))) {
      $path .= '/index.html';
    }
    return $this->getStaticDirectory() . '/' . $this->sanitizePath($path);
  }

  /**
   * Sanitizes a given path by removing leading slashes and query params.
   *
   * @param string $path
   *   The path.
   *
   * @return string
   *   The sanitized path.
   */
  protected function sanitizePath($path) {
    $path = strtok($path, '?');
    return ltrim($path, '/');
  }

  /**
   * Attempts to copy a path from the file system.
   *
   * @param string $path
   *   The path.
   * @param string $destination
   *   The destination.
   *
   * @return bool
   *   TRUE if $path exists and was copied to $destination, FALSE otherwise.
   */
  protected function copyPath($path, $destination) {
    $path = urldecode($path);

    $base_path = base_path();
    if ($base_path !== '/') {
      $base_path = ltrim($base_path, '/');
      $pattern = '|^' . preg_quote($base_path, '|') . '|';
      $path = preg_replace($pattern, '', $path);
    }
    if (file_exists($path)) {
      $directory = dirname($destination);
      file_prepare_directory($directory, FILE_CREATE_DIRECTORY);
      file_unmanaged_copy($path, $destination, FILE_EXISTS_REPLACE);
      return TRUE;
    }
    return FALSE;
  }

  /**
   * Excludes paths provided by system modules.
   *
   * @return string[]
   *   An array of paths that should not be generated.
   */
  protected function getExcludedPaths() {
    $paths = [
      '/big_pipe/no-js',
      '/filter/tips',
      '/history/get_node_read_timestamps',
      '/machine_name/transliterate',
      '/system/files',
      '/system/temporary',
      '/system/timezone',
      '/batch',
      '/session/token',
      '/user/password',
      '/user/login',
      '/user/login_status',
      '/views/ajax',
      '/user/register',
      '/system/4xx',
      '/system/404',
      '/system/403',
      '/system/401',
      '/comments/render_new_comments_node_links',
      '/media/oembed',
    ];
    $site_paths = Settings::get('tome_static_path_exclude', []);
    if (is_array($site_paths)) {
      $paths = array_merge($paths, $site_paths);
    }
    foreach ($paths as &$path) {
      $path = rtrim(base_path(), '/') . $path;
    }
    return $paths;
  }

}
