<?php

namespace Drupal\Tests\tome\Kernel;

use Drupal\Core\Datetime\Entity\DateFormat;
use Drupal\Core\Site\Settings;
use Drupal\node\Entity\Node;

/**
 * Tests that static site generation works.
 *
 * @coversDefaultClass \Drupal\tome\StaticGenerator
 * @group tome
 */
class StaticGeneratorTest extends TestBase {

  /**
   * @covers \Drupal\tome\StaticGenerator::getPaths
   * @covers \Drupal\tome\EventSubscriber\LanguagePathSubscriber::collectPaths
   */
  public function testGetPaths() {
    /** @var \Drupal\tome\StaticGenerator $static */
    $static = \Drupal::service('tome.static');
    $this->assertEquals([
      '/',
      '/french',
    ], $static->getPaths());

    $article = Node::create(['type' => 'article', 'title' => 'My article']);
    $article->addTranslation('fr', ['title' => 'My french article']);
    $article->save();

    $this->assertEquals([
      '/',
      '_entity:node:en:1',
      '_entity:node:fr:1',
      '/french',
    ], $static->getPaths());

    $settings = Settings::getInstance() ? Settings::getAll() : [];
    $settings['tome_static_path_exclude'] = ['_entity:node:fr:1', '/french'];
    new Settings($settings);

    $this->assertEquals([
      '/',
      '_entity:node:en:1',
    ], $static->getPaths());
  }

  /**
   * @covers \Drupal\tome\StaticGenerator::requestPath
   */
  public function testRequestPath() {
    $this->enableModules(['tome_test']);

    /** @var \Drupal\tome\StaticGenerator $static */
    $static = \Drupal::service('tome.static');

    $static->requestPath('/tome-test/page');
    $this->assertContains('Hello, world', file_get_contents(Settings::get('tome_static_directory') . '/tome-test/page/index.html'));
  }

  /**
   * @covers \Drupal\tome\StaticGenerator::requestPath
   * @covers \Drupal\tome\StaticGenerator::getCssAssets
   * @covers \Drupal\tome\StaticGenerator::getHtmlAssets
   * @covers \Drupal\tome\StaticGenerator::exportPaths
   */
  public function testComplexHtml() {
    $this->enableModules(['tome_test']);

    /** @var \Drupal\tome\StaticGenerator $static */
    $static = \Drupal::service('tome.static');

    $invoke_paths = $static->requestPath('/tome-test/complex-page');
    $static->exportPaths($invoke_paths);
    $base_dir = Settings::get('tome_static_directory');
    $this->assertContains('Hello, complex world', file_get_contents("$base_dir/tome-test/complex-page/index.html"));
    $module_path = drupal_get_path('module', 'tome_test');
    $this->assertFileExists("$base_dir/$module_path/css/complex.css");
    $this->assertFileExists("$base_dir/$module_path/js/complex.js");
    $this->assertFileExists("$base_dir/$module_path/assets/druplicon-1.png");
    $this->assertFileExists("$base_dir/$module_path/assets/file with space.png");
    $this->assertFileExists("$base_dir/core/misc/druplicon.png");
  }

  /**
   * @covers \Drupal\tome\StaticGenerator::requestPath
   * @covers \Drupal\tome\StaticGenerator::replacePagerUrl
   * @covers \Drupal\tome\StaticGenerator::handlePagers
   * @covers \Drupal\tome\StaticGenerator::exportPaths
   */
  public function testPagers() {
    $this->enableModules(['tome_test']);

    /** @var \Drupal\tome\StaticGenerator $static */
    $static = \Drupal::service('tome.static');

    $invoke_paths = $static->requestPath('/tome-test/pager-page');
    $this->assertContains('/tome-test/pager-page?page=0', $invoke_paths);
    $this->assertContains('/tome-test/pager-page?page=1', $invoke_paths);
    $this->assertContains('/foobar?page=2', $invoke_paths);
    $base_dir = Settings::get('tome_static_directory');
    $contents = file_get_contents("$base_dir/tome-test/pager-page/index.html");
    $this->assertContains('href="/tome-test/pager-page"', $contents);
    $this->assertContains('href="/tome-test/pager-page/page/2"', $contents);
    $this->assertContains('href="/foobar/page/3"', $contents);
  }

  /**
   * @covers \Drupal\tome\StaticGenerator::getPaths
   * @covers \Drupal\tome\StaticCache::isCached
   */
  public function testCache() {
    $this->enableModules(['tome_test']);

    /** @var \Drupal\tome\StaticGenerator $static */
    $static = \Drupal::service('tome.static');
    /** @var \Drupal\tome\StaticCache $cache */
    $cache = \Drupal::service('tome.static_cache');

    DateFormat::create([
      'pattern' => 'D, m/d/Y - H:i',
      'id' => 'medium',
    ])->save();

    /** @var \Drupal\node\Entity\Node $article */
    $article = Node::create(['type' => 'article', 'title' => 'My article']);
    $article->save();

    $path = '_entity:node:en:' . $article->id();

    $this->assertContains($path, $static->getPaths());

    $static->requestPath($path);

    $cache->commit();

    $this->assertNotContains($path, $static->getPaths(TRUE));

    $article->setTitle('Changed!');
    $article->save();

    $this->assertContains($path, $static->getPaths(TRUE));

    $static->requestPath($path);

    $cache->commit();

    $this->assertNotContains($path, $static->getPaths(TRUE));

    $settings = Settings::getInstance() ? Settings::getAll() : [];
    $settings['tome_static_cache_exclude'] = [$path];
    new Settings($settings);

    $this->assertContains($path, $static->getPaths(TRUE));

    $settings['tome_static_cache_exclude'] = ['/_entity:node:.*/'];
    new Settings($settings);

    $this->assertContains($path, $static->getPaths(TRUE));

    $invoke_paths = $static->requestPath('/tome-test/pager-page');
    $this->assertContains('/tome-test/pager-page?page=0', $invoke_paths);
    $static->requestPath('/tome-test/pager-page?page=0');

    $cache->commit();

    $invoke_paths = $static->requestPath('/tome-test/pager-page', TRUE);
    $this->assertNotContains('/tome-test/pager-page?page=0', $invoke_paths);
  }

}
